/****************************************************************************************************
 * Sample code for validating RESTOP                                                                *
 * An SPI digital accelerometer (ADXL362) is connected to the MCU MSP430FR5739                      *
 *                                                                                                  *
 * Citation. If you are using this code for your research please cite:                              *
 *                                                                                                  *
 * [1] A. Rodriguez Arreola, D. Balsamo, G. V. Merrett, and A. S. Weddell, RESTOP: Retaining       *
 *     External Peripheral State in Intermittently-Powered Sensor Systems, Sensors, vol. 18,       *
 *     no. 1, p. 172, Jan. 2018. doi:10.3390/s18010172                                              *
 *                                                                                                  *
 * For more information about this work, direct to: http://www.mdpi.com/1424-8220/18/1/172          *
 * Other parts of the work can be found at: http://www.transient.ecs.soton.ac.uk/                   *
 * Alberto R.A. University of Southampton. All rights reserved.  15/Jan/2018                        *
 * This software is provided "as is", without warranty of any kind, distributed                     *
 * under the terms and conditions of the Creative Commons Attribution (CC BY) license               *
 * (http://creativecommons.org/licenses/by/4.0/).                                                   *
 ****************************************************************************************************/

#include <msp430.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdbool.h>
#include <math.h>

#include "Config.h"
#include "RESTOP_func.h"

struct node array[Table];          	//Size declared in Config.h
int pos=0;							//Total number of saved instructions
int lastp=0;
uint8_t PRxData;
int first2=0;
uint8_t fpos=0;
uint8_t regis[2];
uint8_t TXByteCtr, Rx = 0;
//uint8_t m_i[(int)( sizeof(pwr_pi) / sizeof(pwr_pi[0]))] = { 0 };	//Num. of saved instructions per peripheral
uint8_t check_node(uint8_t preserve, uint8_t chip1,uint8_t com1,uint8_t regs1,uint8_t value1, bool protocol)  //Substitute repeated node
{

	uint8_t found=0;
	uint8_t a=0;

	while(!found && (a < pos))
	{
		found=((chip1==(array[a].pcs&0x07))&&(com1 == array[a].cmd)&&(regs1 == array[a].reg)&&((array[a].pcs&0x20) != 0x20)&&((array[a].pcs&0xC0) == (preserve&0xC0)));  //same comm, chip, no preserve and config.
		if(!found)
		{
			a++;
		}
	}

	if(found)								//Substitute with new values
	{
		array[a].pcs = preserve+chip1;		//preserve flag and chip select  (preserve<<4|chip1)
		array[a].cmd = com1;
		array[a].reg = regs1;
		array[a].val = value1;
		if (a==first2)
			first2=array[a].next;

		if (array[array[a].next].prev==a && array[a].prev!=a)
			array[array[a].next].prev = array[a].prev;
		else if (array[a].prev==a)
			array[array[a].next].prev = array[a].next;
		array[array[a].prev].next = array[a].next;

		array[lastp].next = a;
		array[a].next = pos;
		array[a].prev = lastp;
		array[a].prot = protocol;


		lastp = a;	//index of the section that points the last cell //

	}

	return(found);
}

void add_instruction (uint8_t preserve, uint8_t chip, uint8_t com,uint8_t regs,uint8_t value, bool protocol)    //Save instruction in the Table (to be improved in Future work)
{
	//preserve
	/*
	 * 'd0: Not-sate
	 * 'd1: Save
	 * 'd2: Save-but-replace
	 * 'd5: Save and Preserve
	 * 'd6: Save-but-replace and Preserve
	 */

	uint8_t f=0;


	if (regs==reg_reset[chip-1])				//RESET
	{

		fpos=0;
		int a1=0;
		while (a1<pos)
		{
			if (chip==(array[a1].pcs&0x07))  	//Bits [2:0]
			{
				if (first2==a1)
				{
					array[array[a1].next].prev = array[a1].next;		//New first cell
					first2 = array[a1].next;
				}
				else if (lastp==a1)
					lastp = array[a1].prev;

				if (a1 != array[a1].prev && (array[array[a1].prev].prev!=0 && array[array[a1].next].next!=0))   //check if it is not a deleted cell
					array[array[a1].prev].next = array[a1].next;
				array[a1].prev = 0;
				array[a1].next = 0;
			}
			a1++;
		}

	}

	if (preserve&0x10==0x10)											//If Save-but-replace
		f=check_node (preserve,chip,com,regs,value,protocol);		//Check if the address was already written. Only update the value. Not new Node needed

	if (!f || (preserve&0x08==0x08))		//The address has not been written before or save but not replace
	{

		int local_pos=fpos;

		while (local_pos<=pos && f==0)
		{
			if ((array[local_pos].next==0 && array[local_pos].prev==0) || local_pos==pos)   //First element
			{
				array[local_pos].pcs = preserve+chip;		//preserve flag and chip select   CS-ID[2:0] PRV[5:3] R[6] B[7]
				array[local_pos].cmd = com;
				array[local_pos].reg = regs;
				array[local_pos].val = value;
				array[local_pos].prot = protocol;

				if (pos==0)
					array[local_pos].prev=0;
				else
					array[local_pos].prev=lastp;
				if (local_pos == pos)
					array[local_pos].next = pos+1;
				else
					array[local_pos].next = pos;
				array[lastp].next = local_pos;
				lastp = local_pos;
				if (local_pos==pos)
					pos++;				//Array Index
				f++;
			}
			local_pos++;
			fpos++;
		}

	}

}

void clock_config(void)
{
	// Initialize clocks
	CSCTL0_H = 0xA5;                          				// Unlock Clock System registers
	// Configure CPU clock for MCLK = 12MHz, SMCLK = 12MHz, ACLK = 32kHz
	//CSCTL1 |= DCORSEL + DCOFSEL0 + DCOFSEL1; 				// Select high DCO range, frequency
	CSCTL1 &= ~DCORSEL;   									//Low_speed
	//CSCTL1 |= DCORSEL;
	CSCTL1 |= DCOFSEL0 + DCOFSEL1;							//11b=8MHz
	CSCTL2 |= SELA__VLOCLK + SELS__DCOCLK + SELM__DCOCLK;	// Select clock sources
	CSCTL3 |= 0x0;		//DIVA_1 + DIVS_1 + DIVM_1;       				// Select clock dividers A/1 S/2 M/2  DIVA_1 */
}
void SPI_init(void)
{
	// Initialize clocks
	clock_config();

	P1OUT |=  BIT3;
	P1DIR |=  BIT3;		//UCB0STE Slave transmit enable  eUSCI_B0 SPI mode
	P4OUT |=  BIT0;
	P4DIR |=  BIT0;		//UCB0STE Slave transmit enable  eUSCI_B0 SPI mode
	P2DIR |=  BIT2;		//UCB0CLK
	P1DIR |=  BIT6;		//UCB0SIMO
	P1DIR &= ~BIT7;		//UCB0SOMI Input
	P1OUT |=  BIT3;		//SLAVE
	/**********************************************************************************************************
	 * UCSWRST --> Put state machine in reset
	 * UCMST   --> Master mode select
	 * UCSYNC  --> Synchronous mode enable
	 * UCCKPH  --> Clock polarity select: 1b = The inactive state is high.
	 * UCMSB   --> MSB first select. Controls the direction of the receive and transmit shift register
	 * UCMODE_2--> 10b = 4-pin SPI with UCxSTE active low: Slave enabled when UCxSTE = 0
	 * UCSTEM  --> 1b = STE pin is used to generate the enable signal for a 4-wire slave
	 **********************************************************************************************************/
	UCB0CTLW0 |= UCSWRST;	// Put state machine in reset
	//UCB0CTLW0 |= UCSWRST | UCSSEL_2;
	UCB0CTLW0 |= UCMST+UCCKPH+UCMSB+UCSYNC;   // 3-pin, 8-bit SPI master
	UCB0CTLW0 |= UCSSEL_2;                    // SMCLK
	//UCB0CTLW0 |= UCCKPH | UCMSB | UCMST | UCSYNC;
	UCB0BR0 = 0x01;             // UCLK/2
	UCB0BR1 = 0;
	P1SEL1 |= BIT6  | BIT7;  	//P1.6=UCB0SIMO; P1.7=UCB0SOMI
	P2SEL1 |= BIT2;				//UCB0CLK
	P1DIR  |= BIT6;				//USCIB0 USCIB0_SIMO
	P2DIR  |= BIT2;				//USCIB0_UCLK
	UCB0CTLW0 &= ~UCSWRST;		//Release from reset and initialize USCI state machine
}
void RESTOP_strobe(uint8_t preserve, uint8_t cs1, uint8_t reg1, bool protocol)
{
	 preserve = preserve<<3;  		//Shift to left to give space for ID
	 preserve=preserve+0x4;			//It's a strobe reg
	 if (preserve!=0x38 || reg1==reg_reset[cs1-1])
		 add_instruction(preserve,cs1,0,reg1,0,protocol);
	 RESTOP_write1(cs1, 0,reg1, 0,1);

}
void RESTOP_write(uint8_t preserve, uint8_t cs1, uint8_t reg1, uint8_t val1, bool burst, bool protocol)
{
	preserve = preserve<<3;  //Shift to left to give space for ID
	if (burst)
		preserve = preserve + 0x80;   //Burst is bit 8
	if ((preserve&0x38)!=0 || reg1==reg_reset[cs1-1])   //Bits [6:4] for Prv
		add_instruction(preserve,cs1,cmd_write[cs1-1],reg1,val1,protocol);
	if (protocol == SPI)
	{
		RESTOP_write1(preserve+cs1, cmd_write[cs1-1], reg1, val1,0);   //RESTOP_write1(cs1, cmd_write[cs1-1], reg1, val1,0);
	}
	else if (protocol == I2C)
	{
		transmitI2C_init(i2c_add[cs1-1]);  //Setup_TX(i2c_add[cs1-1]);
		transmitI2C(reg1, val1,1);
	}

}
void Transmit(void)		//To be updated in newer versions
{
    while (UCB0CTL1 & UCTXSTP);             // Ensure stop condition got sent
    UCB0CTL1 |= UCTR + UCTXSTT;             // I2C TX, start condition
    __bis_SR_register(CPUOFF + GIE);        // Enter LPM0 w/ interrupts
}
void Receive(void)		//To be updated in newer versions
{
        while (UCB0CTL1 & UCTXSTP);             // Ensure stop condition got sent
        UCB0CTL1 &= ~UCTR ;                     // Clear UCTR
        UCB0CTL1 |= UCTXSTT;                    // I2C start condition
        while (UCB0CTL1 & UCTXSTT);             // Start condition sent?
        UCB0CTL1 |= UCTXSTP;                    // I2C stop condition
        __bis_SR_register(CPUOFF + GIE);        // Enter LPM0 w/ interrupts
}
void transmitI2C_init(uint8_t slave_address)  //To be updated in newer versions
{
	P1SEL1 |= BIT6 + BIT7;                  // Pin init SDA=Pin6
	/*UCB0CTLW0 |= UCSWRST;                   // put eUSCI_B in reset state
	UCB0CTLW0 |= UCMODE_3 + UCMST + UCSSEL_2;// I2C master mode, SMCLk
	UCB0BRW = 0x8;                          // baudrate = SMCLK / 8
	UCB0TBCNT = 0x01; 						// TX 2 bytes of data  Register+value or 1 byte: register
	UCB0I2CSA = slave_address;                       // address slave is 48hex
	UCB0CTLW0 &=~ UCSWRST;	            //clear reset register
	UCB0IE |= UCTXIE0 + UCNACKIE;           //transmit and NACK interrupt enable */
	UCB0CTLW0 |= UCSWRST;	            //Software reset enabled
	UCB0CTLW0 = UCSSEL_2+ UCSWRST;
	UCB0CTLW0 |= UCMODE_3  + UCMST + UCSYNC; //I2C mode, Master mode, sync
	//UCB0CTLW1 |= UCASTP_2;		    //automatic stop generated
											//after UCB0TBCNT is reached
	UCB0BRW    = 0x0008; 		    // baudrate = SMCLK / 8
	UCB0TBCNT |= 0x0001;		    //number of bytes to be received
	UCB0I2CSA  = 0x0039;		    //slave address
	UCB0CTL1  &=~UCSWRST;
	UCB0IE    |=  UCTXIE0+UCRXIE+UCNACKIE+UCBCNTIE;
	P1DIR |= BIT0;
}
uint8_t transmitI2C(uint8_t reg1, uint8_t val1, uint8_t oper)  //To be updated in newer versions
{
	while (UCB0CTLW0 & UCTXSTP);            // Ensure stop condition got sent
	if (oper==1)
	{
		TXByteCtr=2;
		regis[0]=reg1;
		regis[1]=val1;
		Transmit();
	}
	else
	{
		TXByteCtr=1;
		regis[0]=reg1;
		Receive();
	}
    //UCB0CTLW0 |= UCTR + UCTXSTT;            // I2C TX, start condition
    //UCB0TXBUF = reg1;
    while (UCB0CTLW0 & UCTXSTP);            // Ensure stop condition got sent
   /* if (oper==1)
    	UCB0TXBUF = val1;
    else
    	return UCB0RXBUF; */
    if (oper==0)
    	return PRxData;

}
void RESTOP_write1(uint8_t preserve, uint8_t cmd1, uint8_t reg1, uint8_t val1, uint8_t strb)   //No driver, used for restoring to reduce time by avoiding "VARARGS". "strb" for strobe inst.
{

	uint8_t read,cs1,burst;
	cs1=preserve&0x07;		//Peripheral ID

	if ((preserve&0x80==0x80 && reg1!=0x00) || (preserve&0x80==0x00))   //It's not a burst or it is the first element of a burst transmission
	{
		chip_select[cs1-1]();
		chip_selectn[cs1-1]();
		burst = 1;
	}
	//while (!(UCB0IFG&UCTXIFG));     // Wait for TXBUF ready Check bus is not busy
	while(UCB0STATW & UCBUSY);
	if (cmd_write[cs1-1]!=0 && burst < 2)
	{
		UCB0TXBUF=cmd1;				//Write or Read?
		burst = 2;
	}
	//while(UCB0STATW & UCBUSY);	//Check bus is not busy
	read = UCB0RXBUF;				//Read to remove Overrun error flag
	while (!(UCB0IFG&UCTXIFG));     // Wait for TXBUF ready Check bus is not busy
	if ((preserve&0x80==0x80 && reg1!=0x00) || (preserve&0x80==0x00))    //It's not a burst or it's the first instruction of a burst transmission.
		UCB0TXBUF=reg1;
	//while(UCB0STATW & UCBUSY);   	//Check bus is not busy
	read = UCB0RXBUF;			  	//Read to remove Overrun error flag
	while (!(UCB0IFG&UCTXIFG));     // Wait for TXBUF ready Check bus is not busy
	if (!strb)						//Not a Strobe instruction
		UCB0TXBUF=val1;				//Value to be set in register
	while(UCB0STATW & UCBUSY);   	//Check bus is not busy
	read = UCB0RXBUF;			 	//Read to remove Overrun error flag

	if (preserve&0x80==0x00)	//preserve&0x80==0x00
	{
		chip_select[cs1-1]();
		burst=0;
	}

}
uint8_t RESTOP_read(uint8_t preserve, uint8_t cs1, uint8_t reg1, bool burst, bool protocol)			//Read function
{
	preserve = preserve<<3;  		//Shift to left to give space for ID
	preserve = preserve + 0x40;		//It's a read instruction Bit 7
	if (burst)
		preserve = preserve + 0x80;   //Burst is bit 8
	if ((preserve&0x38)!=0)   //Bits [6:4] for Prv
		add_instruction(preserve,cs1,cmd_read[cs1-1],reg1,0,protocol);
	if (protocol == SPI)
		return RESTOP_read1(preserve+cs1,cmd_read[cs1-1], reg1);
	else if (protocol == I2C)
	{
		transmitI2C_init(i2c_add[cs1-1]);  //Setup_TX(i2c_add[cs1-1]);
		return transmitI2C(reg1, 0,0);
	}
}
uint8_t RESTOP_read1(uint8_t preserve,uint8_t cmd2,uint8_t reg2)
{
	uint8_t data;
	uint8_t cs2,burst;
	cs2=preserve&0x07;

	if ((preserve&0x80==0x80 && reg2!=0x00) || (preserve&0x80==0x00))   //It's not a burst or first element of a burst transmission
	{
		chip_select[cs2-1]();
		chip_selectn[cs2-1]();
		burst = 1;
	}
	//while (!(UCB0IFG&UCTXIFG));      //Check bus is not busy
	while(UCB0STATW & UCBUSY);
	if ((cmd_write[cs2-1]!=0) && (burst < 2))
	{
		UCB0TXBUF=cmd2;             //Write or Read?
		burst = 2;
	}
	while(UCB0STATW & UCBUSY);		//Check bus is not busy
	data = UCB0RXBUF;
	while (!(UCB0IFG&UCTXIFG));		//Check bus is not busy
	if ((preserve&0x80==0x80 && reg2!=0x00) || (preserve&0x80==0x00))    //It's not a burst or it's the first instruction of a burst transmission.
		UCB0TXBUF=reg2;					//What register to write?
	while(UCB0STATW & UCBUSY);		//Check bus is not busy
	data = UCB0RXBUF;				//Read value
	while (!(UCB0IFG&UCTXIFG));		//Check bus is not busy
	UCB0TXBUF=0;				  	//What register to write?
	while(UCB0STATW & UCBUSY);   	//Check bus is not busy
	data = UCB0RXBUF;
	if (preserve&0x80==0x00)        //No burst, disable periphral
	{
		chip_select[cs2-1]();
		burst = 0;
	}
	return data;
}
void RESTOP_restore(void)
{
	uint8_t cs1,cnt=0;
	uint8_t cell=first2;

	while (cnt<pos)
	{

		if ((array[cell].pcs & 0x40)==0)  	 //WRITE INSTRUCTION
		{
			if (array[cell].prot == SPI)
				RESTOP_write1(array[cell].pcs&0x0F, array[cell].cmd, array[cell].reg, array[cell].val,array[cell].pcs&0x80);
			else
			{
				cs1=array[cell].pcs&0x03;
				transmitI2C_init(i2c_add[cs1-1]);  //Setup_TX(i2c_add[cs1-1]);
				transmitI2C(array[cell].reg, array[cell].val,1);
			}
		}
		else								 //READ INSTRUCTION
		{
			if (array[cell].prot == SPI)
				RESTOP_read1(array[cell].pcs&0x0F, array[cell].cmd, array[cell].reg);
			else
			{
				cs1=array[cell].pcs&0x03;
				transmitI2C_init(i2c_add[cs1-1]);  		//Setup_TX(i2c_add[cs1-1]);
				transmitI2C(array[cell].reg, 0,0);
			}
		}

		cell=array[cell].next;
		cnt++;
		if ((array[cell].prev==0)&& (array[cell].next==0))
			cnt=pos;
	}

}

float RESTOP_VR_adj(void)		//Function to adjust the restore threshold VR
{
	float E_rps = 0.00;		//Energy consumed to restore all peripherals
	uint8_t n_i = (int)( sizeof(pwr_pi) / sizeof(pwr_pi[0]));	//Number of attached peripherals.
	uint8_t j = 0;		//Index of saved instructions
	float T_spi = 0.0;	//Total time SPI
	float T_i2c = 0.0;	//Total time I2C

	//j = m_i[n_i-1];	//Saved instructions of Pi
	while (j<pos)
	{
		if(array[j].prot == SPI)
		{
			if ((cmd_write[array[j].pcs&0x03] != 0) || (cmd_read[array[j].pcs&0x03] != 0))	//Peripheral needs a command
			{
				if (((array[j].pcs & 0x40)==0) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0))  //Write Instruction, no strobe, no burst
					T_spi = T_spi + Time_spi[0];  //3 parameters
				else if ((((array[j].pcs & 0x40)==1) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0)) || ((array[j].pcs & 0x04)==1)) //Read Instruction, no burst or Strobe
					T_spi = T_spi + Time_spi[1];  //2 parameters
				else
					T_spi = T_spi + Time_spi[2];  //1 parameter
			}
			else
			{
				if (((array[j].pcs & 0x40)==0) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0))  //Write Instruction, no strobe, no burst
					T_spi = T_spi + Time_spi[1];  //2 parameters
				else
					T_spi = T_spi + Time_spi[2];  //1 parameter
			}
		}
		else if(array[j].prot == I2C)
		{
			if ((cmd_write[array[j].pcs&0x03] != 0) || (cmd_read[array[j].pcs&0x03] != 0))	//Peripheral needs a command
			{
				if (((array[j].pcs & 0x40)==0) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0))  //Write Instruction, no strobe, no burst
					T_i2c = T_i2c + Time_i2c[0];  //3 parameters
				else if ((((array[j].pcs & 0x40)==1) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0)) || ((array[j].pcs & 0x04)==1)) //Read Instruction, no burst or Strobe
					T_i2c = T_i2c + Time_i2c[1];  //2 parameters
				else
					T_i2c = T_i2c + Time_i2c[2];  //1 parameter
			}
			else
			{
				if (((array[j].pcs & 0x40)==0) && ((array[j].pcs & 0x04)==0) && ((array[j].pcs & 0x80)==0))  //Write Instruction, no strobe, no burst
					T_i2c = T_i2c + Time_i2c[1];  //2 parameters
				else
					T_i2c = T_i2c + Time_i2c[2];  //1 parameter
			}

		}
		j++;
	}
	n_i--;		//0 to n_i-1

	E_rps = pwr_pi[0]*T_spi;			//Total energy to restore peripheral
	if ((n_i>0) && (pwr_pi[n_i]!=0))
		E_rps = E_rps + pwr_pi[n_i-1]*T_i2c;

	return sqrtf(Vmin*Vmin + 2*(0.0000057+E_rps)/Cap);	//New Vr To be improved in Future Work

}

#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector = USCI_B0_VECTOR
__interrupt void USCIB0_ISR(void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(USCI_B0_VECTOR))) USCIB0_ISR (void)
#else
#error Compiler not supported!
#endif /**/
{
  if(Rx == 1)
  {                              // Master Recieve?
      PRxData = UCB0RXBUF;                       // Get RX data
      __bic_SR_register_on_exit(CPUOFF);        // Exit LPM0
  }

  else
  {                                     // Master Transmit
      if (TXByteCtr)                            // Check TX byte counter
        {
          UCB0TXBUF = regis[TXByteCtr-1];                     // Load TX buffer
          TXByteCtr--;                            // Decrement TX byte counter
        }
        else
        {
          //UCB0CTL1 |= UCTXSTP;                    // I2C stop condition
          UCB0CTLW0 |= UCTXSTP;
          //IFG2 &= ~UCB0TXIFG;                     // Clear USCI_B0 TX int flag
          UCB0IFG &= ~UCTXIFG0;
          __bic_SR_register_on_exit(CPUOFF);      // Exit LPM0
        }
  }

}
